#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__version__ = '0.0.8'
__date__ = '06/2015'
__author__ = 'Mathieu Fourcroy'
__contact__ = 'mathieu.fourcroy@gmail.com'
__copyright__ = 'Copyright (c) 2015 Mathieu Fourcroy'
